#include <Arduino.h>
#include <WiFi.h>
#include <string.h>
#include "mbedtls/base64.h"
#include <HTTPClient.h>

#include <WebServer.h>
#include "SPIFFS.h"

#include "Audio.h"
Audio audio;

#define BUFF_MAX      34000

//----------PCM5102 (I2S) ------------
#define   bclk    8   
#define   wclk    17  
#define   dout    18  

// Enter your WiFi setup here:
const char *SSID = "xxxxxxxx";
const char *PASSWORD = "yyyyyyyy";

IPAddress ip(192, 168, 3, 250);           // IP Address
IPAddress gateway(192,168, 3, 1);         // Gateway Address
IPAddress dns(192,168, 3, 1);             // Gateway Address
IPAddress subnet(255, 255, 255, 0);       // Subnet Address

WebServer _server(80);
String _index_top =
                  "<!DOCTYPE html>\n<html>\n<head>\n<meta charset='utf-8'>\n<meta name='viewport' content='width=device-width,initial-scale=1'>\n"
                  "<link rel='stylesheet' type='text/css' href='/radiko.css' >\n<title>Radiko</title>\n</head>\n"
                  "<body>\n<center>\n<div class='b_frame'>\n<br>\n<div class='t_font'><u>Radiko 1.0</u></div>\n<br>\n"
                  "<form method='get' id='abc'>\n<input name='1' id='123' style='display:none'>\n</form>\n"
                  "<form oninput='op.value=ab.value'>\n<span class='v_font'>Volume</span><br>\n"
                  "<input style='width:220px;' type='range' value='";
                  
uint8_t  _st_selected, _station_MAX, _volume;
bool _flg_play, _flg_mute, _flg_auto;
char _url[200], _auth[50];
String  st_list[2], _auth_token;

void setup()
{
   //-----------------  part0 ------------------------------------------------------------------------------
    Serial.begin(115200);
    delay(500);

    WiFi.disconnect();
    WiFi.softAPdisconnect(true);
    delay(500);
    
    WiFi.mode(WIFI_STA);
    WiFi.config(ip, gateway, subnet, dns);
    WiFi.begin(SSID, PASSWORD);
    while (WiFi.status() != WL_CONNECTED)
    {
      Serial.println("...Connecting to WiFi");
      delay(1000);
    }
    Serial.println("Connected");
    Serial.println(SSID);
    Serial.print("IP address: ");
    Serial.println(WiFi.localIP());

    //-----------------  HTTP Server & SPIFFS ----------------------------------------------------------------------
    _server.on("/", handleRoot);
    _server.onNotFound(handleWebRequests);
    _server.begin();
    Serial.println("HTTP server started");

    Serial.println("SPIFFS started");
    SPIFFS.begin();

    Serial.print("I2S Setup");
    audio.setPinout(bclk, wclk, dout);
    audio.setVolumeSteps(100);
    
    Read_state();
    Radiko_Key(_auth_token, st_list);
    
    _flg_mute = false;
    audio.setVolume(_volume); 
    
    if(_flg_auto) {
      _flg_play = true;
      Radiko_play(_st_selected);
    }
    else _flg_play = false;

}

void loop()
{
    _server.handleClient();

    if(_flg_play){
      if (audio.isRunning()) audio.loop();
      else audio.connecttohost(_url,"-", _auth);
    }
}

void handleRoot()
{
  String cmd,str;
  int a,flg;
  
    flg = 1;
    cmd = _server.argName(0);
    switch(cmd.toInt())
    {
      case 1: // Play, Mute, Auto, Save
                cmd = _server.arg("1");
                str = cmd.substring(2);
                a = str.toInt();
                flg = 0;
                switch(cmd.charAt(0))
                {
                  case '1': // volume data
                            _volume = a;
                            audio.setVolume(_volume); 
                            break;
                            
                  case '2': // station No
                            _st_selected = a;
                            Radiko_stop();
                            delay(50);
                            Radiko_play(_st_selected);
                            break;
                            
                  case '3': // play
                            if(!_flg_play) Radiko_play(_st_selected);
                            break;

                  case '4': // mute
                            _flg_mute = !_flg_mute;
                            a = _volume;
                            if(_flg_mute) a = 5; 
                            audio.setVolume(a);
                            break;

                  case '5': // auto flg
                            _flg_auto = !_flg_auto;
                            break;

                  case '6': // Save
                            Save_state();
                            break;
                }
                break;

      case 2: //  Stop
                Radiko_stop();
                break;
    }

    if(flg)
    {
      str = Make_index(str);
      _server.send(200, "text/html",str);
    }
    else _server.send(204, "text/html","");
}

void Radiko_play(uint8_t _st_selected){
  int a;
  String str;
  
    Radiko_Key(_auth_token, st_list);
    str = "https://f-radiko.smartstream.ne.jp/" + get_st_name(_st_selected, st_list[0]) + "/_definst_/simul-stream.stream/playlist.m3u8";
    for(a = 0; a < str.length(); a ++) _url[a] = str[a];
    _url[a] = 0;
    for(a = 0; a < _auth_token.length(); a ++) _auth[a] = _auth_token[a];
    _auth[a] = 0;
    audio.connecttohost(_url,"-", _auth);
    _flg_play = true;  
}

void Radiko_stop()
{
    if(_flg_play) audio.stopSong();
    _flg_play = false;
}

void Radiko_Key(String& _auth_token, String* st_list)
{
  char * auth_key = "bcd151073c03b352e1ef2fd66c32209da9ca0afa" ;
  HTTPClient http;
  int key_length, offset;
  char base64_encoded[32],token_p[32];
  size_t base64_len;
  const char *headerKeys[] = {"X-Radiko-AuthToken", "X-Radiko-KeyLength", "X-Radiko-KeyOffset"};
  String str;
  int a,b,c;
  WiFiClient * stream;
  uint8_t* _sound_buff;

    _sound_buff = (uint8_t*)malloc(BUFF_MAX + 5);
    //-----------------  part1 ------------------------------------------------------------------------------
    http.begin("https://radiko.jp/v2/api/auth1");
    http.addHeader("User-Agent", "esp32/4");
    http.addHeader("Accept", "*/*");
    http.addHeader("X-Radiko-App", "pc_html5");
    http.addHeader("X-Radiko-App-Version", "0.0.1");
    http.addHeader("X-Radiko-User", "dummy_user");
    http.addHeader("X-Radiko-Device", "pc");
    http.collectHeaders(headerKeys, 3);
    Serial.println("part1 = " + String(http.GET()));
    
    offset = String(http.header(headerKeys[2])).toInt();
    key_length = String(http.header(headerKeys[1])).toInt();
    _auth_token = http.header(headerKeys[0]);
    http.end();

    token_p[0] = '\0';
    strncpy(token_p, &auth_key[offset], key_length);
    token_p[key_length] = '\0';
    mbedtls_base64_encode((unsigned char *) base64_encoded, sizeof(base64_encoded),&base64_len, (unsigned char *) token_p, strlen(token_p));

    //-----------------  part2 ------------------------------------------------------------------------------
    http.begin("https://radiko.jp/v2/api/auth2");
    http.addHeader("X-Radiko-AuthToken", _auth_token);
    http.addHeader("X-Radiko-Partialkey", base64_encoded);
    http.addHeader("X-Radiko-User", "dummy_user");
    http.addHeader("X-Radiko-Device", "pc");
    Serial.println("part2 = " + String(http.GET()));
    
    str = http.getString();
    str = str.substring(0,str.indexOf(","));
    http.end();
  
    //-----------------  part3 ------------------------------------------------------------------------------
    //    http.begin("https://radiko.jp/v2/station/list/[ID].xml");
    str = "https://radiko.jp/v2/station/list/" + str + ".xml";
    http.begin(str);
    http.addHeader("X-Radiko-AuthToken", _auth_token);
    Serial.println("part3 = " + String(http.GET()));
      
    b = http.getSize();
    c = 0;
    stream = http.getStreamPtr();
    while(b)
    {
      while(!(a = stream->available())) NOP();
      stream->readBytes(&_sound_buff[c], a);
      b -= a;
      c += a;
    }
    _sound_buff[c] = 0;
    str = String((char *)_sound_buff);    

    b = 0;
    st_list[0] = st_list[1] = "";
    while((a = str.indexOf("<id>")) != -1)
    {
      a +=  4;
      str = &str[a];
      a = str.indexOf("</id>");
      st_list[0] += str.substring(0,a);
      st_list[0] += "\n";
      a = a + 16;
      str = &str[a];
      a = str.indexOf("</name>");
      st_list[1] += str.substring(0,a);
      st_list[1] += "\n";
      b ++;
    }
    _station_MAX = b;
    for(a = 0; a < b; a ++)
      Serial.println("No." + String(a) +" ID: " + get_st_name(a, st_list[0]) + " , name: " + get_st_name(a, st_list[1]));
    Serial.println("station_MAX = " + String(b));
    
    http.end();
    free(_sound_buff);
}

String get_st_name(int num, String str){
  int a;

    if(num > -1){
      a = 0;
      while(num){
        while(str[a] != 0xa) a ++;  
        a ++;
        num --;        
      }
      num = a;
      a ++;
      while(str[a] != 0xa) a ++;
      str = str.substring(num,a);
    }
    else str = "";
    return(str);
}

String Make_index(String str)
{
  int a;
  
    str = String(_volume);
    str += "' id='vol' name='ab' onchange='set_data(1)'>\n<output name='op' class='v_font'>";
    str += String(_volume);
    str += "</output>\n<br>\n<form>\n<select id='_title' class='m_font' onchange='set_data(2)'>\n";
    
    for(a = 0; a < _station_MAX; a ++)
    {
      str += "<option value='";
      str += (String(a) + "'");
      if(a == _st_selected) str += " selected>";
      else str += ">";
      str += get_st_name(a, st_list[1]);
      str += "</option>\n";
    }

    str += "</select>\n</form>\n<br>\n<form method='get'>\n<button type='button' id='_play' onclick='set_data(3)' ";
    if(_flg_play) str += "style ='background: red;'";
    str += ">Play</button>\n<button type='button' id='_mute' onclick='set_data(4)' ";
    if(_flg_mute) str += "style ='background: red;'";
    
    str += ">Mute</button>\n<button type='submit' name='2' >Stop</button>\n<br>\n<span class='v_font' style ='margin-left:5%;'>Auto Start :</span>\n";
    str += "<input type='radio' id='_auto' style ='margin-left:5%;' onclick='set_data(5)'";
    if(_flg_auto) str += " checked ";
    str += "/>\n<button type='button' style ='margin-left:16%;' onclick='set_data(6)'";
    str += ">Save</button>\n</form>\n</div>\n</center>\n<script src='./radiko.js'></script>\n</body>\n</html>";

    str = _index_top + str;
    
    return str;
}

void handleWebRequests()
{
  String dataType = "text/plain";
  String path;
  File dataFile;

    path = _server.uri();
    if(path.endsWith(".txt")) dataType = "text/plain";
    else if(path.endsWith(".css")) dataType = "text/css";
    else if(path.endsWith(".js")) dataType = "application/javascript";
    else if(path.endsWith(".ico")) dataType = "image/html";
    
    dataFile = SPIFFS.open(path.c_str(), "r");
    _server.streamFile(dataFile, dataType);
    dataFile.close();
    delay(5);
}

void Save_state()
{
  String str;
  File dataFile;

    str = String(_volume) + "\n" + String(_st_selected) +"\n" + String(_flg_auto) + "\n";
    dataFile = SPIFFS.open("/state.txt", "w");
    dataFile.print(str);
    dataFile.close();
}

void Read_state()
{
  String str;
  File dataFile;

    dataFile = SPIFFS.open("/state.txt", "r");
  
    str = dataFile.readStringUntil('\n');
    _volume = str.toInt();

    str = dataFile.readStringUntil('\n');
    _st_selected = str.toInt();
    
    str = dataFile.readStringUntil('\n');
    _flg_auto = str.toInt();

    dataFile.close();
}
